/*
   Package types represents all types from the GraphQL specification in code.


   The names of the Go types, whenever possible, match 1:1 with the names from
   the specification.

*/
package types
